# -*- coding: utf-8 -*-

"""
yaspin.base_spinner
~~~~~~~~~~~~~~~~~~~

Spinner class, used to construct other spinners.
"""

from __future__ import absolute_import

from collections import namedtuple


Spinner = namedtuple("Spinner", "frames interval")
default_spinner = Spinner("⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏", 80)



# -*- coding: utf-8 -*-

"""
tests.test_in_out
~~~~~~~~~~~~~~~~~

Checks that all input data is converted to unicode.
And all output data is converted to builtin str type.
"""

import re
import sys
import time

import pytest

from yaspin import Spinner, yaspin
from yaspin.compat import PY2, basestring, builtin_str, str
from yaspin.constants import ENCODING
from yaspin.helpers import to_unicode


def test_input_converted_to_unicode(text, frames, interval, reversal, side):
    sp = Spinner(frames, interval)
    sp = yaspin(sp, text, side=side, reversal=reversal)

    if isinstance(sp._frames, basestring):
        assert isinstance(sp._frames, str)

    if isinstance(sp._frames, (list, tuple)):
        assert isinstance(sp._frames[0], str)

    assert isinstance(sp._text, str)


def test_out_converted_to_builtin_str(text, frames, interval, reversal, side):
    sp = Spinner(frames, interval)
    sp = yaspin(sp, text, side=side, reversal=reversal)

    for _ in range(len(frames)):
        frame = next(sp._cycle)
        out = sp._compose_out(frame)
        assert isinstance(out, builtin_str)


def test_repr(text, frames, interval):
    sp = Spinner(frames, interval)
    sp = yaspin(sp, text)

    assert isinstance(repr(sp), builtin_str)


def test_compose_out_with_color(
    color_test_cases, on_color_test_cases, attrs_test_cases
):
    color, color_exp = color_test_cases
    on_color, on_color_exp = on_color_test_cases
    attrs, attrs_exp = attrs_test_cases

    # Skip non relevant cases
    empty = not color_exp or not on_color_exp or not attrs_exp
    is_exc = any(
        [
            isinstance(color_exp, Exception),
            isinstance(on_color_exp, Exception),
            isinstance(attrs_exp, Exception),
        ]
    )
    if empty or is_exc:
        items = [repr(color_exp), repr(on_color_exp), repr(attrs_exp)]
        pytest.skip("{0} - unsupported case".format(items))

    # Actual test
    sp = yaspin(color=color, on_color=on_color, attrs=attrs)
    assert sp._color == color
    assert sp._on_color == on_color
    assert sp._attrs == set(attrs)

    out = sp._compose_out(frame=u"/")
    assert out.startswith("\r\033")
    assert isinstance(out, builtin_str)


def test_write(capsys, text):
    sp = yaspin()
    sp.write(text)

    out, _ = capsys.readouterr()
    # cleans stdout from _clear_line and \r
    out = out.replace("\r\033[K", "")

    # handle out and text encodings (come out messy in PY2)
    # Under PY2 ``capsys.readouterr`` always produces ``out``
    # of type ``unicode``. Conversion to bytes is required
    # for proper ``out`` and ``text`` comparison.
    if PY2:
        out = out.encode(ENCODING)
        if isinstance(text, str):
            text = text.encode(ENCODING)

    assert isinstance(out, basestring)
    assert out[-1] == "\n"
    if text:
        assert out[:-1] == text


def test_hide_show(capsys, text, request):
    # Setup
    sp = yaspin()
    sp.start()

    # Ensure that sp.stop() will be executed
    def teardown():
        sp.stop()

    request.addfinalizer(teardown)

    #
    # Actual test
    #
    sp.hide()

    # ensure that hidden spinner flag was set
    assert sp._hide_spin.is_set()
    out, _ = capsys.readouterr()

    # ensure that text was cleared with the hide method
    assert out[-4:] == "\r\033[K"

    # properly encode text to unicode if running in PY2
    if PY2:
        text = to_unicode(text).encode(ENCODING)

    # ``\n`` is required to flush stdout during
    # the hidden state of the spinner
    sys.stdout.write("{0}\n".format(text))
    out, _ = capsys.readouterr()

    # cleans stdout from _clear_line and \r
    out = out.replace("\r\033[K", "")

    # handle out and text encodings (come out messy in PY2)
    # Under PY2 ``capsys.readouterr`` always produces ``out``
    # of type ``unicode``. Conversion to bytes is required
    # for proper ``out`` and ``text`` comparison.
    if PY2:
        out = out.encode(ENCODING)
        if isinstance(text, str):
            text = text.encode(ENCODING)

    assert isinstance(out, basestring)
    assert out[-1] == "\n"
    if text:
        assert out[:-1] == text

    sp.show()

    # ensure that hidden spinner flag was cleared
    assert not sp._hide_spin.is_set()
    out, _ = capsys.readouterr()

    # ensure that text was cleared before resuming the spinner
    assert out[:4] == "\r\033[K"


def test_spinner_write_race_condition(capsys):
    # test that spinner text does not overwrite write() contents
    # this generally happens when the spinner thread writes
    # between write()'s \r and the text it actually wants to write

    sp = yaspin(text="aaaa")
    sp.start()
    sp._interval = 0.0
    start_time = time.time()
    while time.time() - start_time < 3.0:
        sp.write("bbbb")
    sp.stop()

    out, _ = capsys.readouterr()
    assert "aaaa" in out  # spinner text is present
    assert "bbbb" in out  # write() text is present
    assert not re.search(r"aaaa[^\rb]*bbbb", out)



# -*- coding: utf-8 -*-

"""
tests.conftest
~~~~~~~~~~~~~~

Tests data.
"""

import signal
import sys

import pytest

from yaspin.compat import iteritems
from yaspin.constants import COLOR_MAP
from yaspin.signal_handlers import default_handler, fancy_handler


frame_cases = [
    # XXX: try byte strings
    # String types
    "",  # empty
    "+x*",  # ascii str
    "⢄⢂⢁⡁⡈⡐⡠",  # non-ascii in str
    u"⢹⢺⢼⣸",  # non-ascii in unicode str
    # Lists
    [],  # List[]
    [b"\xf0\x9f\x8c\xb2", b"\xf0\x9f\x8e\x84"],  # List[bytes]
    [u"🌲", u"🎄"],  # List[unicode]
    ["⢹", "⢺", "⢼", "⣸"],  # List[str], non-ascii
    # Tuples
    (),  # Tuple[]
    (b"\xf0\x9f\x8c\xb2", b"\xf0\x9f\x8e\x84"),  # Tuple[bytes]
    (u"🌲", u"🎄"),  # Tuple[unicode]
    ("⢹", "⢺", "⢼", "⣸"),  # Tuple[str], non-ascii
]


frame_ids = [
    # String types
    "'empty str'",
    "'ascii str'",
    "'non-ascii str'",
    "'non-ascii unicode str'",
    # Lists
    "'List[]'",
    "'List[bytes]'",
    "'List[unicode]'",
    "'List[str] non-ascii'",
    # Tuples
    "'Tuple[]'",
    "'Tuple[bytes]'",
    "'Tuple[unicode]'",
    "'Tuple[str] non-ascii'",
]


text_cases = [
    # XXX: try byte strings
    "",  # empty
    "Loading",  # ascii str
    "ℙƴ☂ℌøἤ",  # non-ascii in str
    u"Загрузка",  # non-ascii in unicode str
]


text_ids = [
    "'empty'",
    "'ascii str'",
    "'non-ascii str'",
    "'non-ascii unicode str'",
]


@pytest.fixture(scope="session")
def interval():
    return 80


@pytest.fixture(scope="session", params=frame_cases, ids=frame_ids)
def frames(request):
    return request.param


@pytest.fixture(scope="session", params=text_cases, ids=text_ids)
def text(request):
    return request.param


@pytest.fixture(scope="session", params=["left", "right"])
def side(request):
    return request.param


@pytest.fixture(
    scope="session", params=[False, True], ids=["default", "reversal"]
)
def reversal(request):
    return request.param


def color_id_func(case):
    if isinstance(case, tuple):
        color, _ = case
    else:
        color = case

    if not color or callable(color):
        val = repr(color)
    else:
        val = color

    return val


def attrs_id_func(case):
    if isinstance(case, list):
        val = ", ".join(case)
    if isinstance(case, tuple):
        attrs, _ = case
        val = ", ".join(attrs)
    return val


@pytest.fixture(
    scope="session",
    ids=color_id_func,
    params=[
        # Empty values
        ("", ""),
        (None, None),
        # Supported text colors
        ("red", "red"),
        ("green", "green"),
        ("yellow", "yellow"),
        ("blue", "blue"),
        ("magenta", "magenta"),
        ("cyan", "cyan"),
        ("white", "white"),
        # Unsupported text colors
        ("Red", ValueError()),
        ("orange", ValueError()),
    ],
)
def color_test_cases(request):
    return request.param


@pytest.fixture(
    scope="session",
    ids=color_id_func,
    params=[
        # Empty values
        ("", ""),
        (None, None),
        # Supported highlights
        ("on_red", "on_red"),
        ("on_green", "on_green"),
        ("on_yellow", "on_yellow"),
        ("on_blue", "on_blue"),
        ("on_magenta", "on_magenta"),
        ("on_cyan", "on_cyan"),
        ("on_white", "on_white"),
        # Unsupported highlights
        ("on_foo", ValueError()),
    ],
)
def on_color_test_cases(request):
    return request.param


@pytest.fixture(
    scope="session",
    ids=attrs_id_func,
    params=[
        # Supported attrs
        (["bold"], ["bold"]),
        (["dark"], ["dark"]),
        (["underline"], ["underline"]),
        (["blink"], ["blink"]),
        (["reverse"], ["reverse"]),
        (["concealed"], ["concealed"]),
        # Multiple attrs
        (["bold", "dark"], ["bold", "dark"]),
        (["bold", "dark", "reverse"], ["bold", "dark", "reverse"]),
        # Unsupported attrs
        (["Dark"], ValueError()),
        (["bold", "bar"], ValueError()),
    ],
)
def attrs_test_cases(request):
    return request.param


@pytest.fixture(
    scope="session",
    ids=color_id_func,
    params=sorted([k for k, v in iteritems(COLOR_MAP) if v == "color"]),
)
def supported_colors(request):
    return request.param


@pytest.fixture(
    scope="session",
    ids=color_id_func,
    params=sorted([k for k, v in iteritems(COLOR_MAP) if v == "on_color"]),
)
def supported_highlights(request):
    return request.param


@pytest.fixture(
    scope="session",
    ids=attrs_id_func,
    params=sorted(
        [[k] for k, v in iteritems(COLOR_MAP) if v == "attrs"]
        + [
            ["bold", "dark"],
            ["blink", "concealed", "reverse"],
            ["underline", "concealed", "bold", "dark", "blink", "reverse"],
        ]
    ),
)
def supported_attrs(request):
    return request.param


@pytest.fixture(
    scope="session",
    params=[
        # Empty
        b"",
        u"",
        # Success
        b"OK",
        u"OK",
        b"\xe2\x9c\x94",
        u"✔",
        # Sun
        b"\xe2\x98\x80\xef\xb8\x8f",
        u"☀️",
        # Spark
        b"\xf0\x9f\x92\xa5",
        u"💥",
    ],
)
def final_text(request):
    return request.param


@pytest.fixture(
    scope="session",
    params=[
        None,
        {signal.SIGUSR1: signal.SIG_DFL},
        {signal.SIGTERM: signal.SIG_IGN},
        {signal.SIGTERM: signal.default_int_handler},
        {signal.SIGHUP: default_handler},
        {signal.SIGINT: fancy_handler},
        {signal.SIGINT: lambda signum, frame: sys.exit(1)},
        {
            signal.SIGUSR1: signal.SIG_DFL,
            signal.SIGTERM: signal.SIG_IGN,
            signal.SIGHUP: default_handler,
            signal.SIGINT: fancy_handler,
            signal.SIGINT: lambda signum, frame: sys.exit(1),
        },
    ],
    ids=[
        "no sigmap",
        "SIGUSR1 - SIG_DFL",
        "SIGTERM - SIG_IGN",
        "SIGTERM - default_int_handler",
        "SIGHUP - default_handler",
        "SIGINT - fancy_handler",
        "SIGINT - custom handler",
        "Multiple signals-handlers map",
    ],
)
def sigmap_test_cases(request):
    return request.param



# -*- coding: utf-8 -*-

"""
tests.test_attrs
~~~~~~~~~~~~~~~~

Test Yaspin attributes magic hidden in __getattr__.
"""

from __future__ import absolute_import

import pytest

from yaspin import yaspin
from yaspin.compat import iteritems
from yaspin.constants import COLOR_MAP, SPINNER_ATTRS
from yaspin.spinners import Spinners


@pytest.mark.parametrize("attr_name", SPINNER_ATTRS)
def test_set_spinner_by_name(attr_name):
    sp = getattr(yaspin(), attr_name)
    assert sp.spinner == getattr(Spinners, attr_name)


# Values for ``color`` argument
def test_color(color_test_cases):
    color, expected = color_test_cases
    # ``None`` and ``""`` are skipped
    if not color:
        pytest.skip("{0} - unsupported case".format(repr(color)))

    sp = yaspin()

    if isinstance(expected, Exception):
        with pytest.raises(AttributeError):
            getattr(sp, color)
    else:
        getattr(sp, color)
        assert sp.color == expected
        assert sp._color_func.keywords["color"] == expected


# Values for ``on_color`` argument
def test_on_color(on_color_test_cases):
    on_color, expected = on_color_test_cases
    # ``None`` and ``""`` are skipped
    if not on_color:
        pytest.skip("{0} - unsupported case".format(repr(on_color)))

    sp = yaspin()

    if isinstance(expected, Exception):
        with pytest.raises(AttributeError):
            getattr(sp, on_color)
    else:
        getattr(sp, on_color)
        assert sp.on_color == expected
        assert sp._color_func.keywords["on_color"] == expected


# Values for ``attrs`` argument
@pytest.mark.parametrize(
    "attr", sorted([k for k, v in iteritems(COLOR_MAP) if v == "attrs"])
)
def test_attrs(attr):
    sp = yaspin()
    getattr(sp, attr)
    assert sp.attrs == [attr]
    assert sp._color_func.keywords["attrs"] == [attr]


def id_func(case):
    return ", ".join(case)


@pytest.mark.parametrize(
    "attrs",
    [
        ["bold", "bold"],
        ["bold", "dark"],
        ["blink", "reverse", "blink"],
        ["concealed", "underline", "bold", "bold"],
    ],
    ids=id_func,
)
def test_previous_attrs_persist_1(attrs):
    sp = yaspin()
    for attr in attrs:
        getattr(sp, attr)
    assert sp._attrs == set(attrs)


def test_previous_attrs_persist_2():
    sp = yaspin(attrs=["blink", "underline"]).bold.dark
    assert sp._attrs == set(["blink", "underline", "bold", "dark"])


# Values for ``side`` argument
@pytest.mark.parametrize(
    "side, expected",
    [("left", "left"), ("right", "right"), ("center", ValueError())],
)
def test_side(side, expected):
    sp = yaspin()

    if isinstance(expected, Exception):
        with pytest.raises(AttributeError):
            getattr(sp, side)
    else:
        getattr(sp, side)
        assert sp.side == expected



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from .__version__ import __version__  # noqa
from .api import kbi_safe_yaspin, yaspin
from .base_spinner import Spinner


__all__ = ("yaspin", "kbi_safe_yaspin", "Spinner")



# -*- coding: utf-8 -*-

"""
examples.right_spinner
~~~~~~~~~~~~~~~~~~~~~~

Left and Right spinner position.
"""

import time

from yaspin import yaspin


def main():
    with yaspin(text="Right spinner", side="right", color="cyan") as sp:
        time.sleep(2)

        # Switch to left spinner
        sp.side = "left"  # or just ``sp.left``
        sp.text = "Left spinner"

        time.sleep(2)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
examples.hide_show
~~~~~~~~~~~~~~~~~~

Basic usage of the ``hide`` and ``show`` methods.
"""

import sys
import time

from yaspin import yaspin


def main():
    with yaspin(text="Downloading images") as sp:
        # task 1
        time.sleep(1)
        sp.hide()
        sys.stdout.write("> image 1 download complete\n")
        sp.show()

        # task 2
        time.sleep(2)
        sp.hide()
        sys.stdout.write("> image 2 download complete\n")
        sp.show()

        # finalize
        sp.green.ok("✔")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
examples.hide_show_prompt_toolkit
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Usage of the ``hide`` and ``show`` methods with prompt_toolkit's printing.

Requires python-prompt-tooklit >= 2.0.1
https://github.com/jonathanslenders/python-prompt-toolkit/

.. code:: bash
    pip install "prompt_toolkit>=2.0.1"
"""

from __future__ import print_function

import sys
import time

from yaspin import yaspin

try:
    from prompt_toolkit import HTML, print_formatted_text
    from prompt_toolkit.styles import Style
except ImportError:
    print(
        "This example requires python-prompt-tooklit >= 2.0.1:\n"
        "https://github.com/jonathanslenders/python-prompt-toolkit/\n"
        "\nTo install it, run:\n"
        'pip install "prompt_toolkit>=2.0.1"'
    )
    sys.exit(1)


# override print with feature-rich ``print_formatted_text`` from prompt_toolkit
print = print_formatted_text

# build a basic prompt_toolkit style for styling the HTML wrapped text
style = Style.from_dict({"msg": "#4caf50 bold", "sub-msg": "#616161 italic"})


with yaspin(text="Downloading images") as sp:

    # task 1
    time.sleep(1)
    sp.hide()
    print(
        HTML(
            u"<b>></b> <msg>image 1</msg> <sub-msg>download complete</sub-msg>"
        ),
        style=style,
    )
    sp.show()

    # task 2
    time.sleep(2)
    sp.hide()
    print(
        HTML(
            u"<b>></b> <msg>image 2</msg> <sub-msg>download complete</sub-msg>"
        ),
        style=style,
    )
    sp.show()

    # finalize
    sp.green.ok("✔")



# -*- coding: utf-8 -*-

"""
examples.write_method
~~~~~~~~~~~~~~~~~~~~~

Basic usage of ``write`` method.
"""

import time

from yaspin import yaspin


def main():
    with yaspin(text="Downloading images") as sp:
        # task 1
        time.sleep(1)
        sp.write("> image 1 download complete")

        # task 2
        time.sleep(2)
        sp.write("> image 2 download complete")

        # finalize
        sp.green.ok("✔")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
examples.demo
~~~~~~~~~~~~~

Yaspin features demonstration.
"""

import random
import time

from yaspin import yaspin
from yaspin.compat import iteritems
from yaspin.constants import COLOR_MAP
from yaspin.spinners import Spinners


COLORS = (k for k, v in iteritems(COLOR_MAP) if v == "color")
HIGHLIGHTS = (k for k, v in iteritems(COLOR_MAP) if v == "on_color")


def any_spinner_you_like():
    params = [
        ("line", 0.8),
        ("dots10", 0.8),
        ("dots11", 0.8),
        ("simpleDots", 1),
        ("star", 0.5),
        ("balloon", 0.7),
        ("noise", 0.5),
        ("arc", 0.5),
        ("arrow2", 0.5),
        ("bouncingBar", 1),
        ("bouncingBall", 1),
        ("smiley", 0.7),
        ("hearts", 0.5),
        ("clock", 0.7),
        ("earth", 0.7),
        ("moon", 0.7),
        ("runner", 0.5),
        ("pong", 1),
        ("shark", 1.5),
        ("christmas", 0.5),
    ]
    random.shuffle(params)

    # Setup printing
    max_len = 45
    msg = "[Any spinner you like]"

    for name, period in params:
        spinner = getattr(Spinners, name)

        spaces_qty = max_len - len(name) - len(msg) - len(spinner.frames[0])
        text = "{0}{1}{2}".format(name, " " * spaces_qty, msg)

        with yaspin(spinner, text=text, color="cyan"):
            time.sleep(period)


def colors_simple(sleep=0.7):
    # Setup printing
    max_len = 25
    msg = "[Colors]"

    with yaspin(Spinners.dots12) as sp:
        for color in COLORS:
            spaces_qty = max_len - len(color) - len(msg)
            text = "{0}{1}{2}".format(color, " " * spaces_qty, msg)

            sp.color = color
            sp.text = text
            time.sleep(sleep)


def color_highlights(sleep=0.5):
    # Setup printing
    max_len = 40
    msg = "[Color Highlights]"

    with yaspin(Spinners.bouncingBall) as sp:
        for highlight in HIGHLIGHTS:
            name = "On {0} color".format(highlight.split("_")[1])
            spaces_qty = max_len - len(name) - len(msg)
            text = "{0}{1}{2}".format(name, " " * spaces_qty, msg)

            sp.on_color = highlight
            sp.text = text

            time.sleep(sleep)


def color_attributes(sleep=0.8):
    descriptions = [
        "Bold white color",
        "Dark red color",
        "Underline green color",
        "Blink yellow color",
        "Reverse blue color",
        "Concealed magenta color",
    ]
    # Setup printing
    max_len = 42
    msg = "[Color Attributes]"

    with yaspin(Spinners.bouncingBall) as sp:
        for descr in descriptions:
            spaces_qty = max_len - len(descr) - len(msg)
            text = "{0}{1}{2}".format(descr, " " * spaces_qty, msg)

            attr, color, _ = descr.split()
            sp.attrs, sp.color = [attr.lower()], color
            sp.text = text

            time.sleep(sleep)


def color_craziness(sleep=1.3):
    descriptions = [
        "Bold underline reverse cyan color",
        "Dark blink concealed white color",
        "Underline red on_grey color",
        "Reversed green on_red color",
    ]

    def parse_attrs(description):
        """Parse attributes from text description."""
        attrs = []
        for word in description.split():
            attr = word.lower()
            sp = yaspin()
            try:
                getattr(sp, attr)
            except AttributeError:
                continue
            else:
                attrs.append(attr)
        return attrs

    # Setup printing
    max_len = 58
    msg = "[Color Craziness 🙀 ]"

    # New spinner instance should be created every iteration since
    # multiple simultaneous color attributes are supported. Hence,
    # updating attribute of the instance will add new attribute to
    # the existing list of previous attributes.
    for descr in descriptions:
        spaces_qty = max_len - len(descr) - len(msg)
        text = "{0}{1}{2}".format(descr, " " * spaces_qty, msg)

        with yaspin(Spinners.pong, text=text) as sp:
            # Apply all color attributes from description
            for attr in parse_attrs(descr):
                getattr(sp, attr)

            time.sleep(sleep)


def right_spinner(sleep=2):
    with yaspin(text="Right spinner", side="right", color="cyan") as sp:
        time.sleep(sleep)

        # Switch to left spinner
        sp.side = "left"
        sp.text = "Left spinner"

        time.sleep(sleep)


def reversed_spinner(sleep=1):
    with yaspin(text="Reversed spinner", reversal=True, color="cyan") as sp:
        time.sleep(sleep)

        sp.spinner = Spinners.line

        time.sleep(sleep)

        sp.text = "Enjoy!"
        sp.ok("☀️ ")


def main():
    any_spinner_you_like()
    colors_simple()
    color_highlights()
    color_attributes()
    color_craziness()
    right_spinner()
    reversed_spinner()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
yaspin.helpers
~~~~~~~~~~~~~~

Helper functions.
"""

from __future__ import absolute_import

from .compat import bytes
from .constants import ENCODING


def to_unicode(text_type, encoding=ENCODING):
    if isinstance(text_type, bytes):
        return text_type.decode(encoding)
    return text_type



# -*- coding: utf-8 -*-

"""
tests.test_yaspin
~~~~~~~~~~~~~~~~~

Basic unittests.
"""

from __future__ import absolute_import

from collections import namedtuple

import pytest

from yaspin import Spinner, yaspin
from yaspin.base_spinner import default_spinner


@pytest.mark.parametrize(
    "spinner, expected",
    [
        # None
        (None, default_spinner),
        # hasattr(spinner, "frames") and not hasattr(spinner, "interval")
        (namedtuple("Spinner", "frames")("-\\|/"), default_spinner),
        # not hasattr(spinner, "frames") and hasattr(spinner, "interval")
        (namedtuple("Spinner", "interval")(42), default_spinner),
        # Both attrs, not set
        (Spinner("", 0), default_spinner),
        # Both attrs, not frames
        (Spinner("", 42), default_spinner),
        # Both attrs, not interval
        (Spinner("-\\|/", 0), default_spinner),
        # Both attrs, are set
        (Spinner("-\\|/", 42), Spinner("-\\|/", 42)),
    ],
    ids=[
        "None",
        "no `interval` attr",
        "no `frames` attr",
        "attrs not set",
        "`frames` not set",
        "`interval` not set",
        "both attrs are set",
    ],
)
def test_set_spinner(spinner, expected):
    sp = yaspin(spinner)
    assert sp.spinner == expected



# -*- coding: utf-8 -*-

"""
tests.test_pipes
~~~~~~~~~~~~~~~~

Checks that scripts with yaspin are safe
for pipes and redirects:

  $ python script_that_uses_yaspin.py > script.log
  $ python script_that_uses_yaspin.py | grep ERROR

"""

import os
import sys
import uuid

import pytest

from yaspin.compat import basestring, str
from yaspin.constants import ENCODING


TEST_CODE = """\
# -*- coding: utf-8 -*-

import time
from yaspin import yaspin, Spinner

with yaspin(Spinner(%s, %s), '%s') as sp:
    time.sleep(0.5)
    sp.fail('🙀')
"""


def to_bytes(str_or_bytes, encoding=ENCODING):
    if isinstance(str_or_bytes, str):
        return str_or_bytes.encode(encoding)
    return str_or_bytes


def to_unicode(str_or_bytes, encoding=ENCODING):
    if isinstance(str_or_bytes, bytes):
        return str_or_bytes.decode(encoding)
    return str_or_bytes


@pytest.fixture()
def case_id():
    return str(uuid.uuid4())


def test_piping_output(text, frames, interval, case_id):
    py_fname = "spin-{0}.py".format(case_id)
    fname = "out-{0}.txt".format(case_id)

    def teardown():
        os.remove(py_fname)
        os.remove(fname)

    with open(py_fname, "wb") as f:
        if isinstance(frames, basestring):
            # enclosing characters in quotes
            # for basestring entries
            # to avoid SyntaxError
            frames = repr(to_unicode(frames))

        text = to_unicode(text)
        interval = to_unicode(interval)
        code = to_unicode(TEST_CODE)
        res = code % (frames, interval, text)
        f.write(to_bytes(res))

    try:
        # $ python spin.py > out.txt
        os.system("{0} {1} > {2}".format(sys.executable, py_fname, fname))
    except UnicodeEncodeError as err:
        pytest.fail(err)
    finally:
        teardown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
yaspin
~~~~~~

A lightweight and easy to use terminal spinner.
No external dependencies.

:copyright: (c) 2018 Pavlo Dmytrenko
:license: MIT
"""

import codecs
import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


# Package meta-data
NAME = "yaspin"
LICENSE = "MIT"
DESCRIPTION = "Yet Another Terminal Spinner"
URL = "https://github.com/pavdmyt/yaspin"
EMAIL = "mail@pavdmyt.com"
AUTHOR = "Pavlo Dmytrenko"


def read_version(package):
    version_path = os.path.join(package, "__version__.py")
    with open(version_path, "r") as fd:
        for line in fd:
            if line.startswith("__version__ = "):
                return line.split()[-1].strip().strip('"')


version = read_version(NAME)


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ["tests/"]
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


with codecs.open("README.rst", encoding="utf-8") as fd:
    readme = fd.read()


tests_require = ["pytest"]


setup(
    name=NAME,
    version=version,
    author=AUTHOR,
    author_email=EMAIL,
    license=LICENSE,
    description=DESCRIPTION,
    long_description=readme,
    url=URL,
    download_url="https://github.com/pavdmyt/yaspin/archive/master.zip",
    packages=find_packages(exclude=("tests", "docs", "examples")),
    include_package_data=True,
    tests_require=tests_require,
    cmdclass={"test": PyTest},
    keywords="progressmeter progress meter rate console terminal console cli"
    " loading loader indicator spinner spinners time busy wait idle",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Other Audience",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: System :: Logging",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Shells",
        "Topic :: Terminals",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

"""
yaspin.yaspin
~~~~~~~~~~~~~

A lightweight terminal spinner.
"""

from __future__ import absolute_import

import functools
import itertools
import signal
import sys
import threading
import time

from .base_spinner import default_spinner
from .compat import PY2, basestring, builtin_str, bytes, iteritems, str
from .constants import COLOR_ATTRS, COLOR_MAP, ENCODING, SPINNER_ATTRS
from .helpers import to_unicode
from .termcolor import colored


class Yaspin(object):
    """Implements a context manager that spawns a thread
    to write spinner frames into a tty (stdout) during
    context execution.
    """

    # When Python finds its output attached to a terminal,
    # it sets the sys.stdout.encoding attribute to the terminal's encoding.
    # The print statement's handler will automatically encode unicode
    # arguments into bytes.
    #
    # In Py2 when piping or redirecting output, Python does not detect
    # the desired character set of the output, it sets sys.stdout.encoding
    # to None, and print will invoke the default "ascii" codec.
    #
    # Py3 invokes "UTF-8" codec by default.
    #
    # Thats why in Py2, output should be encoded manually with desired
    # encoding in order to support pipes and redirects.

    def __init__(
        self,
        spinner=None,
        text="",
        color=None,
        on_color=None,
        attrs=None,
        reversal=False,
        side="left",
        sigmap=None,
    ):
        # Spinner
        self._spinner = self._set_spinner(spinner)
        self._frames = self._set_frames(self._spinner, reversal)
        self._interval = self._set_interval(self._spinner)
        self._cycle = self._set_cycle(self._frames)

        # Color Specification
        self._color = self._set_color(color) if color else color
        self._on_color = self._set_on_color(on_color) if on_color else on_color
        self._attrs = self._set_attrs(attrs) if attrs else set()
        self._color_func = self._compose_color_func()

        # Other
        self._text = self._set_text(text)
        self._side = self._set_side(side)
        self._reversal = reversal

        # Helper flags
        self._stop_spin = None
        self._hide_spin = None
        self._spin_thread = None
        self._last_frame = None
        self._stdout_lock = threading.Lock()

        # Signals

        # In Python 2 signal.SIG* are of type int.
        # In Python 3 signal.SIG* are enums.
        #
        # Signal     = Union[enum.Enum, int]
        # SigHandler = Union[enum.Enum, Callable]
        self._sigmap = sigmap if sigmap else {}  # Dict[Signal, SigHandler]
        # Maps signals to their default handlers in order to reset
        # custom handlers set by ``sigmap`` at the cleanup phase.
        self._dfl_sigmap = {}  # Dict[Signal, SigHandler]

    #
    # Dunders
    #
    def __repr__(self):
        repr_ = u"<Yaspin frames={0!s}>".format(self._frames)
        if PY2:
            return repr_.encode(ENCODING)
        return repr_

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        # Avoid stop() execution for the 2nd time
        if self._spin_thread.is_alive():
            self.stop()
        return False  # nothing is handled

    def __call__(self, fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)

        return inner

    def __getattr__(self, name):
        # CLI spinners
        if name in SPINNER_ATTRS:
            from .spinners import Spinners

            sp = getattr(Spinners, name)
            self.spinner = sp
        # Color Attributes: "color", "on_color", "attrs"
        elif name in COLOR_ATTRS:
            attr_type = COLOR_MAP[name]
            # Call appropriate property setters;
            # _color_func is updated automatically by setters.
            if attr_type == "attrs":
                self.attrs = [name]  # calls property setter
            if attr_type in ("color", "on_color"):
                setattr(self, attr_type, name)  # calls property setter
        # Side: "left" or "right"
        elif name in ("left", "right"):
            self.side = name  # calls property setter
        # Common error for unsupported attributes
        else:
            raise AttributeError(
                "'{0}' object has no attribute: '{1}'".format(
                    self.__class__.__name__, name
                )
            )
        return self

    #
    # Properties
    #
    @property
    def spinner(self):
        return self._spinner

    @spinner.setter
    def spinner(self, sp):
        self._spinner = self._set_spinner(sp)
        self._frames = self._set_frames(self._spinner, self._reversal)
        self._interval = self._set_interval(self._spinner)
        self._cycle = self._set_cycle(self._frames)

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, txt):
        self._text = self._set_text(txt)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = self._set_color(value) if value else value
        self._color_func = self._compose_color_func()  # update

    @property
    def on_color(self):
        return self._on_color

    @on_color.setter
    def on_color(self, value):
        self._on_color = self._set_on_color(value) if value else value
        self._color_func = self._compose_color_func()  # update

    @property
    def attrs(self):
        return list(self._attrs)

    @attrs.setter
    def attrs(self, value):
        new_attrs = self._set_attrs(value) if value else set()
        self._attrs = self._attrs.union(new_attrs)
        self._color_func = self._compose_color_func()  # update

    @property
    def side(self):
        return self._side

    @side.setter
    def side(self, value):
        self._side = self._set_side(value)

    @property
    def reversal(self):
        return self._reversal

    @reversal.setter
    def reversal(self, value):
        self._reversal = value
        self._frames = self._set_frames(self._spinner, self._reversal)
        self._cycle = self._set_cycle(self._frames)

    #
    # Public
    #
    def start(self):
        if self._sigmap:
            self._register_signal_handlers()

        if sys.stdout.isatty():
            self._hide_cursor()

        self._stop_spin = threading.Event()
        self._hide_spin = threading.Event()
        self._spin_thread = threading.Thread(target=self._spin)
        self._spin_thread.start()

    def stop(self):
        if self._dfl_sigmap:
            # Reset registered signal handlers to default ones
            self._reset_signal_handlers()

        if self._spin_thread:
            self._stop_spin.set()
            self._spin_thread.join()

        sys.stdout.write("\r")
        self._clear_line()

        if sys.stdout.isatty():
            self._show_cursor()

    def hide(self):
        """Hide the spinner to allow for custom writing to the terminal."""
        thr_is_alive = self._spin_thread and self._spin_thread.is_alive()

        if thr_is_alive and not self._hide_spin.is_set():
            with self._stdout_lock:
                # set the hidden spinner flag
                self._hide_spin.set()

                # clear the current line
                sys.stdout.write("\r")
                self._clear_line()

                # flush the stdout buffer so the current line
                # can be rewritten to
                sys.stdout.flush()

    def show(self):
        """Show the hidden spinner."""
        thr_is_alive = self._spin_thread and self._spin_thread.is_alive()

        if thr_is_alive and self._hide_spin.is_set():
            with self._stdout_lock:
                # clear the hidden spinner flag
                self._hide_spin.clear()

                # clear the current line so the spinner is not appended to it
                sys.stdout.write("\r")
                self._clear_line()

    def write(self, text):
        """Write text in the terminal without breaking the spinner."""
        # similar to tqdm.write()
        # https://pypi.python.org/pypi/tqdm#writing-messages
        with self._stdout_lock:
            sys.stdout.write("\r")
            self._clear_line()

            _text = to_unicode(text)
            if PY2:
                _text = _text.encode(ENCODING)

            # Ensure output is bytes for Py2 and Unicode for Py3
            assert isinstance(_text, builtin_str)

            sys.stdout.write("{0}\n".format(_text))

    def ok(self, text="OK"):
        """Set Ok (success) finalizer to a spinner."""
        _text = text if text else "OK"
        self._freeze(_text)

    def fail(self, text="FAIL"):
        """Set fail finalizer to a spinner."""
        _text = text if text else "FAIL"
        self._freeze(_text)

    #
    # Protected
    #
    def _freeze(self, final_text):
        """Stop spinner, compose last frame and 'freeze' it."""
        text = to_unicode(final_text)
        self._last_frame = self._compose_out(text, mode="last")

        # Should be stopped here, otherwise prints after
        # self._freeze call will mess up the spinner
        self.stop()
        with self._stdout_lock:
            sys.stdout.write(self._last_frame)

    def _spin(self):
        while not self._stop_spin.is_set():

            if self._hide_spin.is_set():
                # Wait a bit to avoid wasting cycles
                time.sleep(self._interval)
                continue

            # Compose output
            spin_phase = next(self._cycle)
            out = self._compose_out(spin_phase)

            # Write
            with self._stdout_lock:
                sys.stdout.write(out)
                self._clear_line()
                sys.stdout.flush()

            # Wait
            time.sleep(self._interval)

    def _compose_color_func(self):
        fn = functools.partial(
            colored,
            color=self._color,
            on_color=self._on_color,
            attrs=list(self._attrs),
        )
        return fn

    def _compose_out(self, frame, mode=None):
        # Ensure Unicode input
        assert isinstance(frame, str)
        assert isinstance(self._text, str)

        frame = frame.encode(ENCODING) if PY2 else frame
        text = self._text.encode(ENCODING) if PY2 else self._text

        # Colors
        if self._color_func is not None:
            frame = self._color_func(frame)

        # Position
        if self._side == "right":
            frame, text = text, frame

        # Mode
        if not mode:
            out = "\r{0} {1}".format(frame, text)
        else:
            out = "{0} {1}\n".format(frame, text)

        # Ensure output is bytes for Py2 and Unicode for Py3
        assert isinstance(out, builtin_str)

        return out

    def _register_signal_handlers(self):
        # SIGKILL cannot be caught or ignored, and the receiving
        # process cannot perform any clean-up upon receiving this
        # signal.
        if signal.SIGKILL in self._sigmap.keys():
            raise ValueError(
                "Trying to set handler for SIGKILL signal. "
                "SIGKILL cannot be cought or ignored in POSIX systems."
            )

        for sig, sig_handler in iteritems(self._sigmap):
            # A handler for a particular signal, once set, remains
            # installed until it is explicitly reset. Store default
            # signal handlers for subsequent reset at cleanup phase.
            dfl_handler = signal.getsignal(sig)
            self._dfl_sigmap[sig] = dfl_handler

            # ``signal.SIG_DFL`` and ``signal.SIG_IGN`` are also valid
            # signal handlers and are not callables.
            if callable(sig_handler):
                # ``signal.signal`` accepts handler function which is
                # called with two arguments: signal number and the
                # interrupted stack frame. ``functools.partial`` solves
                # the problem of passing spinner instance into the handler
                # function.
                sig_handler = functools.partial(sig_handler, spinner=self)

            signal.signal(sig, sig_handler)

    def _reset_signal_handlers(self):
        for sig, sig_handler in iteritems(self._dfl_sigmap):
            signal.signal(sig, sig_handler)

    #
    # Static
    #
    @staticmethod
    def _set_color(value):
        # type: (str) -> str
        available_values = [k for k, v in iteritems(COLOR_MAP) if v == "color"]

        if value not in available_values:
            raise ValueError(
                "'{0}': unsupported color value. Use one of the: {1}".format(
                    value, ", ".join(available_values)
                )
            )
        return value

    @staticmethod
    def _set_on_color(value):
        # type: (str) -> str
        available_values = [
            k for k, v in iteritems(COLOR_MAP) if v == "on_color"
        ]
        if value not in available_values:
            raise ValueError(
                "'{0}': unsupported on_color value. "
                "Use one of the: {1}".format(
                    value, ", ".join(available_values)
                )
            )
        return value

    @staticmethod
    def _set_attrs(attrs):
        # type: (List[str]) -> Set[str]
        available_values = [k for k, v in iteritems(COLOR_MAP) if v == "attrs"]

        for attr in attrs:
            if attr not in available_values:
                raise ValueError(
                    "'{0}': unsupported attribute value. "
                    "Use one of the: {1}".format(
                        attr, ", ".join(available_values)
                    )
                )
        return set(attrs)

    @staticmethod
    def _set_spinner(spinner):
        if hasattr(spinner, "frames") and hasattr(spinner, "interval"):
            if not spinner.frames or not spinner.interval:
                sp = default_spinner
            else:
                sp = spinner
        else:
            sp = default_spinner

        return sp

    @staticmethod
    def _set_side(side):
        # type: (str) -> str
        if side not in ("left", "right"):
            raise ValueError(
                "'{0}': unsupported side value. "
                "Use either 'left' or 'right'."
            )
        return side

    @staticmethod
    def _set_frames(spinner, reversal):
        # type: (base_spinner.Spinner, bool) -> Union[str, List]
        uframes = None  # unicode frames
        uframes_seq = None  # sequence of unicode frames

        if isinstance(spinner.frames, basestring):
            uframes = to_unicode(spinner.frames) if PY2 else spinner.frames

        # TODO (pavdmyt): support any type that implements iterable
        if isinstance(spinner.frames, (list, tuple)):

            # Empty ``spinner.frames`` is handled by ``Yaspin._set_spinner``
            if spinner.frames and isinstance(spinner.frames[0], bytes):
                uframes_seq = [to_unicode(frame) for frame in spinner.frames]
            else:
                uframes_seq = spinner.frames

        _frames = uframes or uframes_seq
        if not _frames:
            # Empty ``spinner.frames`` is handled by ``Yaspin._set_spinner``.
            # This code is very unlikely to be executed. However, it's still
            # here to be on a safe side.
            raise ValueError(
                "{0!r}: no frames found in spinner".format(spinner)
            )

        # Builtin ``reversed`` returns reverse iterator,
        # which adds unnecessary difficulty for returning
        # unicode value;
        # Hence using [::-1] syntax
        frames = _frames[::-1] if reversal else _frames

        return frames

    @staticmethod
    def _set_interval(spinner):
        # Milliseconds to Seconds
        return spinner.interval * 0.001

    @staticmethod
    def _set_cycle(frames):
        return itertools.cycle(frames)

    @staticmethod
    def _set_text(text):
        if PY2:
            return to_unicode(text)
        return text

    @staticmethod
    def _hide_cursor():
        sys.stdout.write("\033[?25l")
        sys.stdout.flush()

    @staticmethod
    def _show_cursor():
        sys.stdout.write("\033[?25h")
        sys.stdout.flush()

    @staticmethod
    def _clear_line():
        sys.stdout.write("\033[K")



# -*- coding: utf-8 -*-

"""
tests.test_finalizers
~~~~~~~~~~~~~~~~~~~~~

"""

from yaspin import yaspin
from yaspin.compat import builtin_str


def test_freeze(final_text):
    sp = yaspin()
    sp._freeze(final_text)

    assert isinstance(sp._last_frame, builtin_str)
    assert sp._last_frame[-1] == "\n"


def test_ok(final_text):
    sp = yaspin()
    sp.ok(final_text)

    assert isinstance(sp._last_frame, builtin_str)
    assert sp._last_frame[-1] == "\n"


def test_ok_empty_case():
    sp = yaspin()
    sp.ok()
    assert "OK" in sp._last_frame


def test_fail(final_text):
    sp = yaspin()
    sp.fail(final_text)

    assert isinstance(sp._last_frame, builtin_str)
    assert sp._last_frame[-1] == "\n"


def test_fail_empty_case():
    sp = yaspin()
    sp.fail()
    assert "FAIL" in sp._last_frame
